@@ -3,8 +3,8 @@
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
  * Copyright (C) 2017-2023 Osimis S.A., Belgium
- * Copyright (C) 2024-2024 Orthanc Team SRL, Belgium
- * Copyright (C) 2021-2024 Sebastien Jodogne, ICTEAM UCLouvain, Belgium
+ * Copyright (C) 2024-2025 Orthanc Team SRL, Belgium
+ * Copyright (C) 2021-2025 Sebastien Jodogne, ICTEAM UCLouvain, Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Affero General Public License
@@ -386,21 +386,6 @@ static void SubmitJob(OrthancPluginRestOutput* output,
 }
 
 
-static void AddInstance(std::list<std::string>& target,
-                        const Json::Value& instance)
-{
-  std::string id;
-  if (OrthancPlugins::LookupStringValue(id, instance, "ID"))
-  {
-    target.push_back(id);
-  }
-  else
-  {
-    throw Orthanc::OrthancException(Orthanc::ErrorCode_InternalError);
-  }
-}
-
-
 static bool GetSequenceSize(size_t& result,
                             const Json::Value& answer,
                             const std::string& tag,
@@ -556,26 +541,28 @@ static void ParseStowRequest(std::list<std::string>& instances /* out */,
     Json::Value tmpInstances;
     if (OrthancPlugins::RestApiGet(tmpResource, "/instances/" + resource, false))
     {
-      AddInstance(instances, tmpResource);
+      instances.push_back(resource);
       AddResourceForJobContent(resourcesForJobContent, Orthanc::ResourceType_Instance, resource);
     }
     // This was not an instance, successively try with series/studies/patients
     else if ((OrthancPlugins::RestApiGet(tmpResource, "/series/" + resource, false) &&
-              OrthancPlugins::RestApiGet(tmpInstances, "/series/" + resource + "/instances", false)) ||
+              OrthancPlugins::RestApiGet(tmpInstances, "/series/" + resource + "/instances?expand=false", false)) ||
              (OrthancPlugins::RestApiGet(tmpResource, "/studies/" + resource, false) &&
-              OrthancPlugins::RestApiGet(tmpInstances, "/studies/" + resource + "/instances", false)) ||
+              OrthancPlugins::RestApiGet(tmpInstances, "/studies/" + resource + "/instances?expand=false", false)) ||
              (OrthancPlugins::RestApiGet(tmpResource, "/patients/" + resource, false) &&
-              OrthancPlugins::RestApiGet(tmpInstances, "/patients/" + resource + "/instances", false)))
+              OrthancPlugins::RestApiGet(tmpInstances, "/patients/" + resource + "/instances?expand=false", false)))
     {
       if (tmpInstances.type() != Json::arrayValue)
       {
         throw Orthanc::OrthancException(Orthanc::ErrorCode_InternalError);
       }
 
+      AddResourceForJobContent(resourcesForJobContent, Orthanc::StringToResourceType(tmpResource["Type"].asString().c_str()), resource);
+
       for (Json::Value::ArrayIndex j = 0; j < tmpInstances.size(); j++)
       {
-        AddInstance(instances, tmpInstances[j]);
-        AddResourceForJobContent(resourcesForJobContent, Orthanc::StringToResourceType(tmpResource["Type"].asString().c_str()), resource);
+        instances.push_back(tmpInstances[j].asString());
+        AddResourceForJobContent(resourcesForJobContent, Orthanc::ResourceType_Instance, tmpInstances[j].asString());
       }
     }
     else
